@@ -20,6 +20,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
+#ident "$Id: insmod.c,v 1.4 1997/12/01 00:46:05 rth Exp $"
 
 #include <sys/types.h>
 #include <stdlib.h>
@@ -30,6 +31,7 @@
 #include <ctype.h>
 #include <errno.h>
 #include <stddef.h>
+#include <getopt.h>
 #include <sys/utsname.h>
 #include <sys/stat.h>
 
@@ -85,6 +87,7 @@ search_module_path(char *base)
     "/lib/modules/%s/block:"
     "/lib/modules/%s/cdrom:"
     "/lib/modules/%s/ipv4:"
+    "/lib/modules/%s/ipv6:"
     "/lib/modules/%s/misc:"
     "/lib/modules/default/fs:"
     "/lib/modules/default/net:"
@@ -92,6 +95,7 @@ search_module_path(char *base)
     "/lib/modules/default/block:"
     "/lib/modules/default/cdrom:"
     "/lib/modules/default/ipv4:"
+    "/lib/modules/default/ipv6:"
     "/lib/modules/default/misc:"
     "/lib/modules/fs:"
     "/lib/modules/net:"
@@ -99,6 +103,7 @@ search_module_path(char *base)
     "/lib/modules/block:"
     "/lib/modules/cdrom:"
     "/lib/modules/ipv4:"
+    "/lib/modules/ipv6:"
     "/lib/modules/misc";
 
   char *path, *p, *filename;
@@ -536,7 +541,55 @@ old_process_module_arguments(struct obj_file *f, int argc, char **argv)
 
       loc = (int *)(f->sections[sym->secidx]->contents + sym->value);
 
-      if (*q >= '0' && *q <= '9')
+      /* Do C quoting if we begin with a ".  */
+      if (*q == '"')
+	{
+	  char *r, *str;
+
+	  str = alloca(strlen(q));
+	  for (r = str, q++; *q != '"'; ++q, ++r)
+	    {
+	      if (*q == '\0')
+		{
+		  error("improperly terminated string argument for %s", p);
+		  return 0;
+		}
+	      else if (*q == '\\')
+		switch (*++q)
+		  {
+		  case 'a': *r = '\a'; break;
+		  case 'b': *r = '\b'; break;
+		  case 'e': *r = '\033'; break;
+		  case 'f': *r = '\f'; break;
+		  case 'n': *r = '\n'; break;
+		  case 'r': *r = '\r'; break;
+		  case 't': *r = '\t'; break;
+
+		  case '0': case '1': case '2': case '3':
+		  case '4': case '5': case '6': case '7':
+		    {
+		      int c = *q - '0';
+		      if (q[1] >= '0' && q[1] <= '7')
+			{
+			  c = (c * 8) + *++q - '0';
+			  if (q[1] >= '0' && q[1] <= '7')
+			    c = (c * 8) + *++q - '0';
+			}
+		      *r = c;
+		    }
+		  break;
+
+		  default:
+		    *r = *q;
+		    break;
+		  }
+	      else
+		*r = *q;
+	    }
+	  *r = '\0';
+	  obj_string_patch(f, sym->secidx, sym->value, str);
+	}
+      else if (*q >= '0' && *q <= '9')
 	{
 	  do
 	    *loc++ = strtol(q, &q, 0);
@@ -959,6 +1012,7 @@ new_process_module_arguments(struct obj_file *f, int argc, char **argv)
 		      else
 			*r = *q;
 		    }
+		  *r = '\0';
 		  ++q;
 		}
 	      else
@@ -998,7 +1052,7 @@ new_process_module_arguments(struct obj_file *f, int argc, char **argv)
 		{
 		  /* Normal string */
 		  obj_string_patch(f, sym->secidx, loc-contents, str);
-		  loc += sizeof(char*);
+		  loc += tgt_sizeof_char_p;
 		}
 	      else
 		{
@@ -1009,8 +1063,8 @@ new_process_module_arguments(struct obj_file *f, int argc, char **argv)
 		  /* Probably we should do that outside the loop ? */
 		  if(!isdigit(*(p + 1)))
 		    {
-		      error("parameter type 'c' for %s must be followed by the maximum size",
-			    key);
+		      error("parameter type 'c' for %s must be followed by"
+			    " the maximum size", key);
 		      return 0;
 		    }
 		  charssize = strtoul(p + 1, (char **) NULL, 10);
@@ -1038,15 +1092,15 @@ new_process_module_arguments(struct obj_file *f, int argc, char **argv)
 		  break;
 		case 'h':
 		  *(short *)loc = v;
-		  loc += sizeof(short);
+		  loc += tgt_sizeof_short;
 		  break;
 		case 'i':
 		  *(int *)loc = v;
-		  loc += sizeof(int);
+		  loc += tgt_sizeof_int;
 		  break;
 		case 'l':
 		  *(long *)loc = v;
-		  loc += sizeof(long);
+		  loc += tgt_sizeof_long;
 		  break;
 
 		default:
@@ -1099,7 +1153,7 @@ new_create_this_module(struct obj_file *f, const char *m_name)
 {
   struct obj_section *sec;
 
-  sec = obj_create_alloced_section_first(f, ".this", sizeof(long),
+  sec = obj_create_alloced_section_first(f, ".this", tgt_sizeof_long,
 					 sizeof(struct new_module));
   memset(sec->contents, 0, sizeof(struct new_module));
 
@@ -1124,7 +1178,7 @@ new_create_module_ksymtab(struct obj_file *f)
       struct new_module_ref *dep;
       struct obj_symbol *tm;
 
-      sec = obj_create_alloced_section(f, ".kmodtab", sizeof(void*),
+      sec = obj_create_alloced_section(f, ".kmodtab", tgt_sizeof_void_p,
 				       (sizeof(struct new_module_ref)
 					* n_ext_modules_used));
       if (!sec)
@@ -1147,7 +1201,7 @@ new_create_module_ksymtab(struct obj_file *f)
       size_t nsyms;
       int *loaded;
 
-      sec = obj_create_alloced_section(f, "__ksymtab", sizeof(void*), 0);
+      sec = obj_create_alloced_section(f, "__ksymtab", tgt_sizeof_void_p, 0);
 
       /* We don't want to export symbols residing in sections that
 	 aren't loaded.  There are a number of these created so that
@@ -1166,16 +1220,16 @@ new_create_module_ksymtab(struct obj_file *f)
 		&& (sym->secidx >= SHN_LORESERVE
 		    || loaded[sym->secidx]))
 	      {
-		ElfW(Addr) ofs = nsyms * 2*sizeof(void*);
+		ElfW(Addr) ofs = nsyms * 2*tgt_sizeof_void_p;
 
 		obj_symbol_patch(f, sec->idx, ofs, sym);
-		obj_string_patch(f, sec->idx, ofs+sizeof(void*), sym->name);
+		obj_string_patch(f, sec->idx, ofs+tgt_sizeof_void_p, sym->name);
 
 		nsyms++;
 	      }
 	}
 
-      obj_extend_section(sec, nsyms * 2 * sizeof(char*));
+      obj_extend_section(sec, nsyms * 2 * tgt_sizeof_char_p);
     }
 
   return 1;
@@ -1200,7 +1254,7 @@ new_init_module(const char *m_name, struct obj_file *f, unsigned long m_size)
   if (sec && sec->header.sh_size)
     {
       module->syms = sec->header.sh_addr;
-      module->nsyms = sec->header.sh_size / (2 * sizeof(char*));
+      module->nsyms = sec->header.sh_size / (2 * tgt_sizeof_char_p);
     }
 
   if (n_ext_modules_used)
@@ -1242,6 +1296,29 @@ new_init_module(const char *m_name, struct obj_file *f, unsigned long m_size)
 
 /*======================================================================*/
 
+/* Print the usage message. */
+static void
+usage(void)
+{
+  fputs("Usage:\n"
+	"insmod [-fkmopsvVxX] [-o name] module [[sym=value]...]\n"
+	"\n"
+	"  module              Filename of a loadable kernel module (*.o)\n"
+	"  -f, --force         Force loading under wrong kernel version\n"
+	"  -k, --autoclean     Make module autoclean-able\n"
+	"  -m                  Generate load map (so crashes can be traced)\n"
+	"  -o NAME\n"
+	"      --name=NAME     Set internal module name to NAME\n"
+	"  -p, --poll          Poll mode; check if the module matches the kernel\n"
+	"  -s, --syslog        Report errors via syslog\n"
+	"  -v, --verbose       Verbose output\n"
+	"  -V, --version       Show version\n"
+	"  -x                  Do not export externs\n"
+	"  -X                  Do export externs (default)\n"
+	, stderr);
+  exit(1);
+}
+
 int
 main(int argc, char **argv)
 {
@@ -1249,6 +1326,16 @@ main(int argc, char **argv)
   int k_crcs;
   int k_new_syscalls;
   char k_strversion[STRVERSIONLEN];
+  static struct option long_opts[] = {
+    { "force", 0, 0, 'f' },
+    { "autoclean", 0, 0, 'k' },
+    { "name", 1, 0, 'o' },
+    { "poll", 0, 0, 'p' },
+    { "syslog", 0, 0, 's' },
+    { "verbose", 0, 0, 'v' },
+    { "version", 0, 0, 'V' },
+    { 0, 0, 0, 0 }
+  };
 
   char *m_name = NULL;
   int m_version;
@@ -1267,7 +1354,8 @@ main(int argc, char **argv)
 
   /* Process the command line.  */
 
-  while ((o = getopt(argc, argv, "fkmno:psvVxX")) != EOF)
+  while ((o = getopt_long(argc, argv, "fkmno:psvVxX",
+			  &long_opts[0], NULL)) != EOF)
     switch (o)
       {
       case 'f': /* force loading */
@@ -1304,28 +1392,13 @@ main(int argc, char **argv)
 	flag_export = 1;
 	break;
       default:
-	goto usage;
+	usage();
+	break;
       }
 
   if (optind >= argc)
     {
-    usage:
-      fputs("Usage:\n"
-	    "insmod [-fkmopsvVxX] [-o name] module [[sym=value]...]\n"
-	    "\n"
-	    "  module     Filename of a loadable kernel module (*.o)\n"
-	    "  -f         Force loading under wrong kernel version\n"
-	    "  -k         Make module autoclean-able\n"
-	    "  -m         Generate load map (so crashes can be traced)\n"
-	    "  -o name    Set internal modulname to name\n"
-	    "  -p         Poll mode; check if the module matches the kernel\n"
-	    "  -s         Report errors via syslog\n"
-	    "  -v         Verbose output\n"
-	    "  -V         Show version\n"
-	    "  -x         do not export externs\n"
-	    "  -X         do export externs (default)\n"
-	    , stderr);
-      return 1;
+      usage();
     }
 
   filename = argv[optind++];
